package org.artifactory.rest.common.service.admin.xray;

import org.apache.commons.lang.StringUtils;
import org.apache.http.HttpStatus;
import org.artifactory.addon.AddonsManager;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.repo.XrayDescriptor;
import org.artifactory.rest.common.model.xray.XrayConfigModel;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.artifactory.security.AuthenticationHelper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

/**
 * @author Chen Keinan
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class UpdateXrayConfigService implements RestService {
    private static final Logger log = LoggerFactory.getLogger(UpdateXrayConfigService.class);

    @Autowired
    private AddonsManager addonsManager;

    @Autowired
    private CentralConfigService configService;

    @Override
    public void execute(ArtifactoryRestRequest request, RestResponse response) {
        if (!addonsManager.isXrayLicensed()) {
            response.error("Invalid xray license for Xray").responseCode(HttpStatus.SC_BAD_REQUEST);
        } else {
            // create xray Config
            if (!updateXrayConfig(request)) {
                response.error("Could not create Xray config - check Artifactory logs").responseCode(400);
            }
        }
    }

    /**
     * update xray config
     */
    private boolean updateXrayConfig(ArtifactoryRestRequest request) {
        String remoteAddress = AuthenticationHelper.getRemoteAddress(AuthenticationHelper.getAuthentication());
        log.info("Updating Xray config, request received from instance at: {}", remoteAddress);
        XrayConfigModel xrayModel = (XrayConfigModel) request.getImodel();
        if (xrayModel.validate(false)) {
            MutableCentralConfigDescriptor descriptor = configService.getMutableDescriptor();
            log.debug("Updating Xray config for instance {}", xrayModel.getXrayBaseUrl());
            XrayDescriptor xrayConfig = descriptor.getXrayConfig();
            updateXrayConfig(xrayConfig, xrayModel.toDescriptor());
            descriptor.setXrayConfig(xrayConfig);
            configService.saveEditedDescriptorAndReload(descriptor);
            return true;
        } else {
            log.debug("Invalid Xray config model received!");
            return false;
        }
    }

    private void updateXrayConfig(XrayDescriptor xrayConfig, XrayDescriptor newXrayConfig) {
        xrayConfig.setEnabled(newXrayConfig.isEnabled());
        xrayConfig.setXrayId(newXrayConfig.getXrayId());
        xrayConfig.setArtifactoryId(newXrayConfig.getArtifactoryId());
        xrayConfig.setBaseUrl(newXrayConfig.getBaseUrl());
        xrayConfig.setUser(newXrayConfig.getUser());
        if (StringUtils.isNotBlank(newXrayConfig.getPassword())) {
            xrayConfig.setPassword(newXrayConfig.getPassword());
        }
    }
}
